/*
 * pwm_softblinker.h
 *
 *  Created on: 22. juni 2020
 *      Author: teig
 */

#ifndef PWM_SOFTBLINKER_H_
    #define PWM_SOFTBLINKER_H_

    #ifndef LIB_PWM_SOFTBLINKER_PARAMS_H_
        #error Define a "_lib_pwm_softblinker_params.h" file in the project that uses this library
    #endif

    // Background info at C++CSP csp::Barrier Class Reference
    // https://www.cs.kent.ac.uk/projects/ofa/c++csp/doc/classcsp_1_1_barrier.html
    typedef enum {synch_none = 0, synch_active = 1} synch_e; // Must be {0,1} like this!

    // Usage in tasks that need synchronization:
    //   Use "blocking_chan_barrier_doynchronize" (below)
    //   Restart timer if necessary
    // ONE send of any value, then one wait on the same channel. This is deadlock free, and contains no state.
    // However, it will lock out any interface calls in the tasks that use this synchronization. But then
    // it's those tasks' clients that have ordered this synchronization.

    // See "Using a chanend in both directions" at
    // https://www.teigfam.net/oyvind/home/technology/141-xc-is-c-plus-x/#using_a_chanend_in_both_directions

    // Observe that a chan is typeless and is therefore deadlock free if sent and received messages are of equal size
    // Also, sending and receiving on the chanend must be in phase

    // 2May2021 (in pwm_softblinker.h in lib_pwm_softblinker)
    typedef enum {  // Div by 4 ok  // Including zero
        steps_0012          =   12, //   13 values
        steps_0100          =  100, //  101 values
        steps_0256          =  256, //  257 values (*)
        steps_1000          = 1000, // 1001 values
        NUM_INTENSITY_STEPS = 4 // Those above
    } intensity_steps_e;

    // (*) Almost like 8-bit grayscale intensity or 8-bit for each of RGB for 24-bit pixels, or 8-bit alpha-channel transparence
    //     Even if a PWM "intensity" is not a continous current but pulse-width modulation
    //     0..256 gives 257 values, but I use it because 256 division by 4 is possible

    #define INTENSITY_STEPS_LIST {steps_0012, steps_0100, steps_0256, steps_1000}

    #define NUM_INTENSITY_STEPS_SHORT (NUM_INTENSITY_STEPS-1)
    #define INTENSITY_STEPS_LIST_SHORT {steps_0012, steps_0100, steps_0256} // NUM_INTENSITY_STEPS_SHORT

    // 2May2021 (in pwm_softblinker.h in lib_pwm_softblinker)
    #define DEFAULT_INTENSITY_STEPS steps_1000

    #define DEFAULT_DARK_INTENSITY  0
    #define DEFAULT_FULL_INTENSITY  DEFAULT_INTENSITY_STEPS
    #define DEFAULT_SYNCH           synch_none // From "barrier.h"

    typedef unsigned intensity_t; // [DEFAULT_DARK_INTENSITY..intensity_steps_e]

    typedef enum {scan_none, scan_continuous}         scan_type_e;
    typedef enum {active_high = 0, active_low = 1}    port_pin_sign_e; // Must be {0,1} like this! Use of XOR is dependent on it!
    typedef enum {continuous_LED, dark_LED, full_LED} start_LED_at_e;
    typedef enum {is_anywhere, is_min, is_max}        extremals_e;

    typedef enum {
        slide_transition_pwm, // PWM pulses will slide with respect to period pulse like yellow_DIRCHANGE
        lock_transition_pwm   // PWM pulses are locked --"--
    } transition_pwm_e;

    #define DEFAULT_TRANSITION_PWM slide_transition_pwm

    // 2May2021 (in PWM_SOFTBLINKER.h in lib_pwm_softblinker)
    #define DEFAULT_PWM_FREQUENCY_HZ 222
    //                               222 Hz no flickering (Should cause no "unperceived neurological effects"). Same as for my aquarium
    //                               100 Hz quite nice
    //                                60 Hz shows the effect quite well
    //                                30 Hz terrible

    #define SOFTBLINK_PERIOD_MIN_MS   (200) //   200 ms (5 blinks per second)
    #define SOFTBLINK_PERIOD_MAX_MS (10000) // 10000 ms

    #define DEFAULT_SOFTBLINK_PERIOD_MS SOFTBLINK_PERIOD_MAX_MS

    // 2May2021 (in PWM_SOFTBLINKER.h in lib_pwm_softblinker)
    #define PERIOD_MS_LIST { \
                SOFTBLINK_PERIOD_MAX_MS,       /* 10000 */ \
                SOFTBLINK_PERIOD_MAX_MS / 2,   /*  5000 */ \
                SOFTBLINK_PERIOD_MAX_MS / 10,  /*  1000 */ \
                SOFTBLINK_PERIOD_MIN_MS * 2.5, /*   500 */ \
                SOFTBLINK_PERIOD_MIN_MS}       /*   200 */

    #define PERIOD_MS_LIST_LEN 5 // Above

    typedef interface softblinker_if {

        bool // max_intensity >= min_intensity
        set_LED_intensity_range ( // FIRST THIS..
                const unsigned          frequency_Hz,    // 0 -> actives port
                const intensity_steps_e intensity_steps, // [1..]
                const intensity_t       min_intensity,   // [0..x]
                const intensity_t       max_intensity);  // [x..intensity_steps_]

        void
        set_LED_period_linear_ms ( // ..THEN THIS
                const unsigned         period_ms, // (*)
                const start_LED_at_e   start_LED_at,
                const transition_pwm_e transition_pwm,
                const synch_e          do_multipart_synch); // do_multipart_synch_not_used in softblinker_task

        // (*) The period goes for any full DARK to FULL (INTENSITY STEPS) BUT IS NORMALISED TO ACTUAL RANGE!
        //     As the range is decreased, the time it takes to deliver out all port output decreases. Example:
        //     A full period of 20 seconds increases for 10 and decreases for 10 seconds. If it starts
        //     at 20% up and stops at 20% from the top its period is not longer 20 seconds, but 12 seconds.
        //     Following the same curve it would start 2 seconds later and reach the top 2 second earlier,
        //     taking 6 seconds to increase. Going down would also take 6 seconds. The sum = 12 seconds.
        //     We therefore normalise to set period

    } softblinker_if;

    // XMOS not raised TICKET (as of 23Jun2020): No matter how much I tweeaked the code, if this was set as
    // a parameter into any of the tasks in pwm_softblinker, I had to touch that file to have it recompiled
    // See code 0178, 0179, 0180, 0181
    #define PWM_PORT_PIN_SIGN active_high // active_low when LED pulled down from 3Vs
                                          // active_high when LED or LED strip drivedn by buffer transistor

    #define SET_LED_INTENSITY_CONTINUOUS_MODE 0

    typedef interface pwm_if {

        void set_LED_intensity (
                const unsigned          frequency_Hz, // 0 -> actives port
                const intensity_steps_e intensity_steps,
                const intensity_t       intensity,    // Normalised to intensity_steps (allways ON when intensity == intensity_steps)
                const transition_pwm_e  transition_pwm);
    } pwm_if;

    #define DO_PULSE_ON_START_SYNCH 1 // 0 or 1

    // [[combinable]] compiles, but [[combine]] par not accepted, so won't be built as combinable
    void barrier_do_chan_task (
            chanend             c_barrier[CONFIG_NUM_SOFTBLIKER_LEDS], // Send and receive id_task_t
            out buffered port:1 outP_external_blue_led_high); // led_on_high_t

    [[combinable]]
    void softblinker_task (
            const id_task_t       id_task,
            client pwm_if         if_pwm,
            server softblinker_if if_softblinker,
            out buffered port:1   out_port_toggle_on_direction_change); // Toggle when LED max

    // [[combinable]] compiles, but [[combine]] par not accepted, so won't be built as combinable
    void softblinker_task_chan_barrier (
            const id_task_t       id_task,
            client pwm_if         if_pwm,
            server softblinker_if if_softblinker,
            out buffered port:1   out_port_toggle_on_direction_change, // Toggle when LED max
            chanend               c_barrier);

    // TODO This controls the DUTY CYCLE
    // TODO There is no deadband between the two LEDs, no period that removes overlap, not even when synchronized
    [[combinable]]
    void pwm_for_LED_task (
            const unsigned      id_task, // For printing only
            server pwm_if       if_pwm,
            out buffered port:1 out_port_LED);  // LED

#else
    #error Nested include PWM_SOFTBLINKER_H_
#endif /* PWM_SOFTBLINKER_H_ */
